// =========================================================================
// Program Name : VesaLib.c
// Author       : Bill Buckels
// Date         : September 25, 1995
// Purpose      : Core Graphics Library Routines for VESA 256 Color Mode
//
// =========================================================================
// Description  : 1. VesaLib.c is written in Large Model MSC 6.00a.
//                   Stack Checking is turned off.
//
//                2. The source is grouped into the following areas:
//                2.1 Global Data
//                2.2 Vesa Initialization
//                2.3 Pixel Graphics
//                2.4 Raster Graphics and BMP routines
//                2.5 Raster Fonts
//
// =========================================================================
// Licence      : This source code is distributed under a Royalty Free
//                Copyright, under the terms and conditions listed below.
//                This Source Code is NOT in The Public Domain.
//
//                1. VesaLib is Copyright Bill Buckels 1995
//                1.1 Bill Buckels is The Author.
//                1.2 This authorship does not apply to public domain
//                    algorithms and methodology that have been used in
//                    this source code.
//                1.3 This authorship does not apply to parts of this
//                    source code that are not owned by The Author,
//                    (i.e. the Windows BMP and ZSOFT PCX file structures).
//                1.4 Rights not specifically stated are reserved by
//                    The Author.
//                1.5 Rights that are specifically stated will not be
//                    changed, except as required by law.
//
//                2. You may use this source code and the resulting
//                   programs for any purpose, commercial or
//                   otherwise, without fee or royalty, under the
//                   following conditions and restrictions:
//                2.1 The Author must be acknowledged as the author.
//                    Nobody else may claim that they wrote or own
//                    this source code.
//                2.1.1 The Acknowledgement of The Author does not need
//                      to be made assertively, as long as nobody else
//                      accepts the credit for authorship.
//                2.1.2 A lack of Acknowledgement must not imply in any
//                      way that The Author is not the author.
//                2.2 This Copyright notice and Licence Agreement
//                    must not be removed or changed.
//                2.3 The Author is not liable for anything.
//                2.4 This Source Code and Its resulting programs must
//                    not be used as part of a warfare system, nor used
//                    in any other way that is not peaceful.
//                2.5 This source code by itself, and its resulting object
//                    code by itself cannot be sold. It must be sold as part
//                    of a product and not as a product itself.
//                    A further restriction also applies... (2.5.1)
//                2.5.1 This Source Code must not be sold as part of a
//                      Graphics Library until a Library Licence is
//                      approved by The Author:
//
//                      Bill Buckels
//                      589 Oxford Street
//                      Winnipeg, Manitoba, Canada R3M 3J2
//                      Voice - (204) 489-1405
//
//                2.5.2 This source code and the demo programs may be
//                      bundled with ANY product (i.e. as sample programs)
//                      under the above restrictions.
//
// =========================================================================
// General Notes:
//
//  1. In addition to the VESALIB source code, it is helpful and almost
//     necessary to obtain the VESA Super VGA BIOS Extensions disk.
//
//     The strategy for VESA and a great deal of other information
//     including drivers and sample programs is included on the VESA disk.
//
//     The following is taken from the VESA disk.
//     The Voice number is correct.
//
//   1.1 Video Electronics Standards Association
//       2150 North First Street, Suite 440
//       San Jose, CA 95131-2029
//
//       Voice (408) 435-0333
//       Fax   (408) 435-8225
//
//  2. Gary Lorenson from Everex has made a large and excellent
//     contribution to the VESA disk and his work should be reviewed
//     by anyone seriously interested in implementing VESA support into
//     an application.
//
// =========================================================================

// =========================================================================
//                2.1 Global Data
// =========================================================================

#define VESALIB

#include "vesalib.h"

// =========================================================================
//                2.2 Vesa Initialization
// =========================================================================

// Is A VESA BIOS Available?
int GetVgaInfo()
{
    vinregs.h.ah = VESA;          // Super Vga Support
    vinregs.h.al = 0;             // Return Super Vga Information

    hl.l = (ulong)&vib;  // Pointer To VGA Information Block
    vsegregs.es  = hl.x[WSEG];
    vinregs.x.di = hl.x[WOFF];

    int86x(VIDEO,&vinregs,&voutregs,&vsegregs);

    if(voutregs.h.al !=VESA)return INITFAIL;
    if(voutregs.h.ah !=SUCCESS)return EXECFAIL;

return SUCCESS;
}

// Get The Information For The VESA Mode...
int GetModeInfo(uint wVesaMode)
{
     vinregs.h.ah = VESA;
     vinregs.h.al = 1;             // Return Super VGA Mode Information
     vinregs.x.cx = wVesaMode;     // Super VGA video mode

     hl.l = (ulong)&mib;          // Pointer to Mode Information Block
     vsegregs.es  = hl.x[WSEG];
     vinregs.x.di = hl.x[WOFF];

     int86x(VIDEO,&vinregs,&voutregs,&vsegregs);

     if(voutregs.h.al !=VESA)return INITFAIL;
     if(voutregs.h.ah !=SUCCESS)return EXECFAIL;

return SUCCESS;
}

// Return The Desired Graphics Vesa Mode if True
// Otherwise Return False...
uint FindVesaMode(uint xres, uint vres,uint noplanes, int pixelbits)
{
   uint wResult;
   int *iVesaMode;

   iVesaMode=vib.VideoModePtr;
   wResult = FALSE;

   for(;;)
   {
     if(*iVesaMode==-1)break;
     if(GetModeInfo(*iVesaMode)==SUCCESS)
     {
     // extended graphics mode...
     if( mib.ModeAttributes &0x02  &&  mib.ModeAttributes &0x10  &&
         mib.XResolution  == xres  &&  mib.YResolution == vres   &&
         mib.NumberOfPlanes == noplanes  &&  mib.BitsPerPixel == pixelbits)
         {
           wResult = *iVesaMode;
           break;
         }
      }
      *iVesaMode++;
   }

return wResult;
}

uint SetVesaMode(uint wVesaMode)
{

    vinregs.h.ah = VESA;
    vinregs.h.al = 2;             // Set Vesa Mode
    vinregs.x.bx = wVesaMode;

    int86(VIDEO,&vinregs,&voutregs);

    vinregs.h.ah = VESA;
    vinregs.h.al = 3;             // Get Current Vesa Mode

    int86(VIDEO,&vinregs,&voutregs);

    // return false if the mode failed to set
    // otherwise return the mode
    if(wVesaMode != voutregs.x.bx)wVesaMode = FALSE;

return wVesaMode;

}

uint GetVesaMode()
{
     vinregs.h.ah = VESA;
     vinregs.h.al = 3;             // Get Current Mode

     int86(VIDEO,&vinregs,&voutregs);

return voutregs.x.bx;
}

void SetVesaBank(uint wWin, uint wBank)
{
    vinregs.h.ah = VESA;
    vinregs.h.al = 5;
    vinregs.h.bh = 0;
    vinregs.h.bl = wWin;
    vinregs.x.dx = wBank;
    int86(VIDEO,&vinregs,&voutregs);
    gib.wSaveBank = wBank;
}

// Put all the global session initialization here
int InitVesaSession()
{

    int i, j;

    // the memory is allocated globally
    // and will be freed by dos when the application ends

    if((gib.lpScanLine = malloc(MAXSCANBUF)) == NULL)return NOMEM;

    for(i=0;i<256;i++)
    {
        for(j=0;j<3;j++)
        {
           gib.aOldRgbInfo[i][j]=0;
           if(i<16)gib.aRgbInfo[i][j]=VgaPalette[i][j]<<2;
           else gib.aRgbInfo[i][j]=0;
        }
    }

    gib.wWhite      = 15;
    gib.wBlack      = 0;
    gib.dwBankSize  = (long )mib.WinSize;
    gib.dwBankSize *= 1024;
    gib.wSaveBank   = 0xffff;
    gib.wReadBank   = 0x00;
    gib.wWriteBank  = 0x00;
    hl.x[WSEG]      = VGASEG;
    hl.x[WOFF]      = 0;
    gib.wXorigin    = 0;
    gib.wYorigin    = 0;

    if (mib.WinAAttributes & 0x05 == 0x05)
      {
          hl.x[WSEG] = mib.WinASegment;
      }
    else
    {
        if (mib.WinBAttributes & 0x05 == 0x05)
        {
          gib.wWriteBank = 0x01;
          hl.x[WSEG]     = mib.WinBSegment;
        }
     }

    gib.lpCrt = (uchar far *)hl.l;


    if (mib.WinAAttributes & 0x03 != 0x03 &&
        mib.WinBAttributes & 0x03 == 0x03)gib.wReadBank = 0x01;


return SUCCESS;

}

void LoadPalette()
{
  int i,j,change;
  int lightest, darkest;

  // don't reset the palette if it hasn't changed
  change = FALSE;
  for(i=0;i<256;i++)
  {
    for(j=0;j<3;j++)
    {
        if(gib.aRgbInfo[i][j]!=gib.aOldRgbInfo[i][j])
        {
            change = TRUE;break;
        }
    }
    if(change==TRUE)break;
  }

  // if the palette has changed reset the palette
  // after we have done so, save the blackest and whitest colors
  // in case we want to draw in the current palette, etc.
  if(change == TRUE)
  {

    if(wClsFlag==TRUE)ClearScreen(0);

    vinregs.h.ah = 0x10;  // function 10h
    vinregs.h.al = 0x12;  // subfunction 12h - set block of color registers
    vinregs.x.bx = 0;     // start with this reg
    vinregs.x.cx = 256;   // do this many
    hl.l = (ulong)&gib.aRgbInfo[0][0];
    vinregs.x.dx = hl.x[WOFF];               // offset to array
    vsegregs.es  = hl.x[WSEG];               // segment to array
    int86x(0x10,&vinregs,&voutregs,&vsegregs);
    lightest = 0;
    darkest  = 256 + 256 + 256;

    for(i=0;i<256;i++)
    {
      change = 0;
      for(j=0;j<3;j++)
      {

        gib.aOldRgbInfo[i][j]=gib.aRgbInfo[i][j];
        change += gib.aRgbInfo[i][j];
      }
      if(change>lightest)
      {
        gib.wWhite=i;lightest=change;
      }
      if(change<darkest)
      {
        gib.wBlack=i;darkest=change;
      }
    }
  }
}

// =========================================================================
//                2.3 Pixel Graphics
// =========================================================================

// return the address of the pixel at x,y
// switch the read and write banks as necessary
unsigned char far *PixAddr(uint x,uint y)
{
   ulong dwTemp, dwOffset, dwPacket;
   uint wOffset, wBank;

   dwTemp =   (long)y;
   dwTemp *=  mib.BytesPerScanLine;
   dwTemp +=  x;

   dwOffset = dwTemp%gib.dwBankSize;
   dwTemp=dwTemp/gib.dwBankSize;
   wBank = (uint )dwTemp;

   // if we don't need to change banks, don't bother
   if(gib.wSaveBank!=wBank)
   {
     SetVesaBank(gib.wWriteBank,wBank);
     SetVesaBank(gib.wReadBank,wBank);
   }
   return &gib.lpCrt[dwOffset];
}

unsigned char ReadPixel(uint x, uint y)
{
    uchar far *ptr;
    ptr = PixAddr(x,y);

return *ptr;
}

void WritePixel(uint x, uint y, uint color)
{
    uchar far *ptr;

    // this will be called in line drawing routines
    // and font writing routines etc. so we
    // need to use a clipping algorithm
    if(x<mib.XResolution && y< mib.YResolution)
    {
        ptr  = PixAddr(x,y);
        *ptr = color;
    }

}

// vertical line
void VLine(int y1, int y2, int x, unsigned color)
{
    int y;

    if(y2<y1)
    {
        y=y1;y1=y2;y2=y;
    }
    y2++;
    for(y=y1;y<y2;y++)WritePixel(x,y,color);

}

// horizontal line
void HLine(int x1, int x2, int y, unsigned color)
{
    int x,i;
    unsigned char far *ptr;

    if(x2<x1)
    {
        x=x1;x1=x2;x2=x;
    }
    x2++;
    i=0;
    ptr = (char *)&gib.lpScanLine[0];
    for(x=x1;x<x2;x++)
    {
     ptr[i]=color;
     i++;
    }
    // avoid recalculating every pixel on a horizontal term...
    VesaRaster(ptr,x1,y,i,WRITEFLAG);
}

// DrawLine Uses Bresenham's Algorithm To Draw Lines
// Unless Straight. In that case it maps to HLine or VLine
int DrawLine(int x1,int y1,int x2,int y2,unsigned color)
{
    #define sign(x) ((x)>0?1:((x)==0?0:(-1)))
    int dx,dy,dxabs,dyabs,i,px,py,sdx,sdy,x,y;

    if(x2==x1)
    {
        VLine(y1,y2,x1,color);return 1;
    }

    if(y2==y1)
    {
        HLine(x1,x2,y1,color);return 2;
    }

    // swap the coordinates if out of order
    if(y2<y1)
    {
        y=y1;y1=y2;y2=y;
    }
    if(x2<x1)
    {
        x=x1;x1=x2;x2=x;
    }

    dx=x2-x1;
    dy=y2-y1;
    sdx=sign(dx);
    sdy=sign(dy);
    dxabs=abs(dx);
    dyabs=abs(dy);
    x=0;
    y=0;
    px=x1;
    py=y1;

    if(dxabs >= dyabs)
    {
        for(i=0;i<=dxabs;i++)
        {
            y+=dyabs;
            if(y>=dxabs)
            {y-=dxabs;
             py+=sdy;
            }
            WritePixel(px,py,color);
            px+=sdx;
        }
    }
    else{
        for(i=0;i<=dyabs;i++)
        {
            x+=dxabs;
            if(x>=dyabs)
            {
                x-=dyabs;
                px+=sdx;
            }
            WritePixel(px,py,color);
            py+=sdy;
        }
    }

return 0;
}

// Draw an Unfilled Box
void DrawBox(int x1, int y1, int x2, int y2,unsigned color)
{
    HLine(x1,x2,y1,color);
    y1++;
    y2--;
    VLine(y1,y2,x2,color);
    VLine(y1,y2,x1,color);
    y2++;
    HLine(x1,x2,y2,color);
}

// get integral aspect ratio for proportionate objects
// based on vertical ratio over horizontal proportion
void GetAspect(int *x, int *y)
{
    int hres, vres;

    // default to 640 x 480, 800 x 600, 1024 x 768, etc.
    // standard aspect ratio is 1:1
    // hres/4 = vres/3
    x[0]=2;
    y[0]=2;

    hres = 1000L * mib.XResolution;
    vres = 1000L * mib.YResolution;

    if(hres/8 == vres/5)
    {
        // 320 x 200, 640 x 400
        x[0]=12;
        y[0]=10;

    }
    else
    {
        if(hres/64 == vres/35)
        {
        // 640 x 350
        x[0]=48;y[0]=35;
        }
    }
}

// plot filled or unfilled circles
// using 8 point symmetry with adjustment for aspect
// using -1 for either color or fillcolor disables that part of this
void DrawCircle(int cx,int cy,int r,int color, int fillcolor)
{

    int x,y,a,b,c,d,e;
    int xaspect,yaspect,points=0;

    GetAspect(&xaspect,&yaspect);

    if(fillcolor>-1)
    {
      x=r; y=0; b=1; e=0; a=(-2)*x+1;
      for(;;)
      {
        c=x*xaspect/yaspect;  d=y*xaspect/yaspect;
        HLine(-c+cx, c+cx,  y+cy,fillcolor);
        HLine(-d+cx, d+cx,  x+cy,fillcolor);
        HLine(-d+cx, d+cx, -x+cy,fillcolor);
        HLine(-c+cx, c+cx, -y+cy,fillcolor);

        if(b>= -a)break;
        y+=1; e+=b; b+=2;
        if(e>r)
        {
          e+=a;a+=2;x-=1;
        }
      }
    }
    if(color>-1)
    {
      x=r; y=0; b=1; e=0; a=(-2)*x+1;
      for(;;)
      {
        c=x*xaspect/yaspect;  d=y*xaspect/yaspect;
        WritePixel( c+cx, y+cy,color);
        WritePixel( d+cx, x+cy,color);
        WritePixel(-d+cx, x+cy,color);
        WritePixel(-c+cx, y+cy,color);
        WritePixel(-c+cx,-y+cy,color);
        WritePixel(-d+cx,-x+cy,color);
        WritePixel( d+cx,-x+cy,color);
        WritePixel( c+cx,-y+cy,color);
        points+=8;
        if(b>= -a)break;
        y+=1; e+=b; b+=2;
        if(e>r)
        {
          e+=a;a+=2;x-=1;
        }
      }
    }
}

// =========================================================================
//                2.4 Raster Graphics
// =========================================================================

// For Microsoft and Borland and other compilers that support the
//   _asm keyword, the following function provides a simple alternative
//   far memory move using string instructions.
// If you need to or want to use memcpy instead, go ahead...

#ifdef GENERIC_C
  #define MoveBytes memcpy
#else
void MoveBytes(char far *lpTarget,char far *lpSource,uint wBytes)
{

#define dest     6
#define src      10
#define nbytes   14

    _asm{

        push    ds
        push    es
        push    di
        push    si

        lds     si, DWORD PTR [bp+src]
        les     di, DWORD PTR [bp+dest]
        mov     cx, WORD PTR  [bp+nbytes]
        cld
        rep     movsb

        pop     si
        pop     di
        pop     es
        pop     ds
       }

#undef dest
#undef src
#undef nbytes
}
#endif

void VesaRaster(uchar *lpszBuf, uint x, uint y, uint wPixels, uint wFLAG)
{

   ulong dwTemp, dwOffset, dwPacket;
   uint wOffset, wBank;

   dwTemp =   (long)y;
   dwTemp *=  mib.BytesPerScanLine;
   dwTemp +=  x;

   dwOffset = dwTemp%gib.dwBankSize;
   dwPacket = dwOffset + wPixels;

   dwTemp=dwTemp/gib.dwBankSize;
   wBank = (uint )dwTemp;

   // if we don't need to change banks, don't bother
   // the only time this will get in the way is if other methods
   // are used to write to the screen between our calls.
   // if this happends the banks can change.
   if(gib.wSaveBank!=wBank)
   {
     SetVesaBank(gib.wWriteBank,wBank);
     SetVesaBank(gib.wReadBank,wBank);
   }

   if(dwPacket<gib.dwBankSize)
   {
      if(wFLAG==READFLAG)
         MoveBytes(&lpszBuf[0],&gib.lpCrt[dwOffset],wPixels);
      else
         MoveBytes(&gib.lpCrt[dwOffset],&lpszBuf[0],wPixels);

   }
   else
   {
      dwPacket  = gib.dwBankSize-dwOffset;
      wOffset = (uint )dwPacket;
      if(wFLAG==READFLAG)
         MoveBytes(&lpszBuf[0],&gib.lpCrt[dwOffset],wOffset);
      else
         MoveBytes(&gib.lpCrt[dwOffset],&lpszBuf[0],wOffset);
      wBank++;
      SetVesaBank(gib.wWriteBank,wBank);
      SetVesaBank(gib.wReadBank,wBank);
      wPixels-=wOffset;
      if(wFLAG==READFLAG)
         MoveBytes(&lpszBuf[wOffset],&gib.lpCrt[0],wPixels);
      else
         MoveBytes(&gib.lpCrt[0],&lpszBuf[wOffset],wPixels);

    }

}

void FBox(uint x1, uint y1, uint x2, uint y2, uchar color)
{
  // create a wiper and use it to draw a filled box
  uint packet,i;

  gib.wSaveBank = 0xffff;            // reset the bank
  packet =(x2-x1)+1;
  for(i=0;i<packet;i++)
    gib.lpScanLine[i]=color;
  y2++;
  for(i=y1;i<y2;i++)
     VesaRaster((char *)&gib.lpScanLine[0],x1,i,packet,WRITEFLAG);
}

// clear screen - draw filled box
void ClearScreen(uint color)
{
    FBox(0,0,mib.XResolution-1,mib.YResolution-1,(unsigned char)color);
}

// translate a 16 color line  to an svga line
void Vga2Svga(unsigned char *src, unsigned char *dest,unsigned width)
{
     int i,j;

     j=0;
     for(i=0;i<width;i++)
     {
        dest[j] = src[i]>>4;  j++;
        dest[j] = src[i]&0xf; j++;
     }
}

// translate a 2 color line to an svga line
void Mono2Svga(unsigned char *src, unsigned char *dest,unsigned width)
{
     int i,j,k;
     uchar msk[]={0x80,0x40,0x20,0x10,0x8,0x4,0x2,0x1};

     j=0;
     for(i=0;i<width;i++)
     {

        for(k=0;k<8;k++)
        {
            if(src[i]&msk[k])dest[j] = gib.wFg;
            else dest[j] = gib.wBg;
            j++;
        }
     }
}

// load 256 color, 16 color, or 2 color Windows BMP files
int LoadVesaBitMap(char *lpszBmpFile)
{

    FILE *fp;
    int i;
    ulong dwSeekSize;
    uint wHres,wVres,packet,wXoff,wYoff;

    if((fp=fopen(lpszBmpFile,"rb"))==NULL)return INITFAIL;

    // read the header stuff into the appropriate structures
    fread((char *)&BitMapFileHeader.bfType,
	             sizeof(BITMAPFILEHEADER),1,fp);
    fread((char *)&bmp.bmiHeader.biSize,
                 sizeof(BITMAPINFO),1,fp);

    if(bmp.bmiHeader.biCompression!=BI_RGB    ||
       bmp.bmiHeader.biWidth >mib.XResolution ||
       bmp.bmiHeader.biHeight>mib.YResolution )
       {
        fclose(fp); return EXECFAIL;
       }

    uif.iFormat = FALSE;
    if(bmp.bmiHeader.biPlanes==1 && bmp.bmiHeader.biBitCount==8)
    {
       uif.iFormat        = SVGA;
       uif.iByteBoundary  = 4 * PIXEL;
       uif.iPalColors     = SVGA;
       uif.iBufOff        = FALSE;
       uif.iDivisor       = PIXEL;
    }
    if(bmp.bmiHeader.biPlanes==1 && bmp.bmiHeader.biBitCount==4)
    {
       uif.iFormat       = VGA;
       uif.iByteBoundary = 8 * PIXEL;
       uif.iPalColors    = VGA;
       uif.iBufOff       = MAXVGABUF;
       uif.iDivisor      = 2 * PIXEL;
    }
    if(bmp.bmiHeader.biPlanes==1 && bmp.bmiHeader.biBitCount==1)
    {
       uif.iFormat       = MONO;
       uif.iByteBoundary = 32 * PIXEL;
       uif.iPalColors    = FALSE;
       uif.iBufOff       = MAXVGABUF;
       uif.iDivisor      = 8 * PIXEL;
    }

    if(uif.iFormat == FALSE)
    {
        fclose(fp); return EXECFAIL;
    }

    for(i=0;i<uif.iPalColors;i++)
    {
       gib.aRgbInfo[i][0]=bmp.bmiColors[i].rgbRed  >>2;
       gib.aRgbInfo[i][1]=bmp.bmiColors[i].rgbGreen>>2;
       gib.aRgbInfo[i][2]=bmp.bmiColors[i].rgbBlue >>2;
    }
    LoadPalette();

    if(uif.iFormat==MONO)
    {
       gib.wFg = bmp.bmiColors[0].rgbRed   +
                bmp.bmiColors[0].rgbGreen + bmp.bmiColors[0].rgbBlue;
       gib.wBg = bmp.bmiColors[1].rgbRed   +
                bmp.bmiColors[1].rgbGreen + bmp.bmiColors[1].rgbBlue;

       if(gib.wBg > gib.wFg)
          { gib.wFg = gib.wWhite; gib.wBg = gib.wBlack; }
       else
          { gib.wBg = gib.wWhite; gib.wFg = gib.wBlack; }
    }

    wHres = (uint )bmp.bmiHeader.biWidth;
    packet = wHres;
    while(wHres%uif.iByteBoundary!=0)wHres++;
    wVres = (uint )bmp.bmiHeader.biHeight;
    wXoff = (mib.XResolution - wHres)/2;
    wYoff = (mib.YResolution - wVres)/2;

    // allow the user over-ride to display windows at
    // specified locations within the view port.
    // unfortunately the BMP file has no load point.
    if(gib.wXorigin!=0xffff && (wHres+gib.wXorigin)<mib.XResolution)
      wXoff = gib.wXorigin;
    if(gib.wYorigin!=0xffff && (wVres+gib.wYorigin)<mib.YResolution)
      wYoff = gib.wYorigin;

    dwSeekSize =  0L;
    dwSeekSize += BitMapFileHeader.bfOffBits;

    // our palette will not always be a full 256 colors...
    // even in some 256 color images

    gib.wSaveBank = 0xffff;            // reset the bank
    fseek(fp,dwSeekSize,SEEK_SET);
    for(i=wVres;i>0;i--)
    {
      fread((char *)&gib.lpScanLine[0],wHres/uif.iDivisor,1,fp);
      switch(uif.iFormat)
       {
         case MONO:  Mono2Svga((char *)&gib.lpScanLine[0],
                               (char *)&gib.lpScanLine[uif.iBufOff],
                               wHres/uif.iDivisor);
                               break;
         case VGA:   Vga2Svga((char *)&gib.lpScanLine[0],
                               (char *)&gib.lpScanLine[uif.iBufOff],
                               wHres/uif.iDivisor);
        }
        VesaRaster((char *)&gib.lpScanLine[uif.iBufOff],wXoff,i-1+wYoff,
                      packet,WRITEFLAG);
    }
    fclose(fp);
    gib.wXorigin = 0;
    gib.wYorigin = 0;

return SUCCESS;
}

// =========================================================================
//                2.5 Raster Fonts
// =========================================================================

void FreeVesaFont()
{
    if(wFontFlag!=0)free(lpFontData);
    wFontFlag=0;
}

int LoadVesaFont(char *lpszFontName)
{
    FILE *fp;
    char *lpszFreeFont = "FREEFONT";
    int i;

    FreeVesaFont();

    if((fp = fopen(lpszFontName,"rb"))==NULL)return -1;

    fread((char *)&fhb,sizeof(FontHeaderBlock),1,fp);
    for(i=0;i<8;i++)
    {
        if(lpszFreeFont[i]!=fhb.szFileType[i])
        {
            fclose(fp);
            return -2;
        }
    }

    if((lpFontData= malloc(fhb.wBufSize))==NULL)
    {
        fclose(fp);
        return -3;
    }

    wFontFlag = 1;
    fseek(fp,fhb.dwHeadSize,SEEK_SET);
    fread(lpFontData,fhb.wBufSize,1,fp);
    fclose(fp);
    return 0;
}

int DrawVesaFont(uchar *lpszStr,
                 int iXorigin,  int iYorigin,
                 char cJustify, uint wColor)
{
    int iScanLine,iByte,iChar,iBit,iTarget;
    int fh,iX,iY,iWidth,iXFactor;
    int iXOff = (fhb.wWidth/8)+1, iYOff, iScanOff;
    unsigned char far *ptr;
    if(wFontFlag!=1)return -1;

    iX=0;
    for(;;)
    {
        if(lpszStr[iX]==13    || lpszStr[iX]==10||
           lpszStr[iX]=='\x1a'|| lpszStr[iX]==0)
        {
           lpszStr[iX]=0;break;
        }
        iX++;
    }
    iTarget=strlen(lpszStr);
    if(iTarget<1)return -2;

    if(!(fhb.wWidth%8))iXOff-=1;
    iYOff = ((fhb.wHeight)*(iXOff))+1;

    // strip hard returns
    // justification
    switch(toupper(cJustify))
        {
        case 'M':
        case 'C': iXorigin-=(((iTarget*fhb.wWidth)/2)+iXOff);
                  iYorigin-=(fhb.wHeight/2);
        default : break;
        }

    // don't bother to clip, but allow equal bytes before and after
    iScanOff = (MAXSCANBUF - gib.wDefaultWidth)/2;
    ptr = (char *)&gib.lpScanLine[iScanOff];
    for (iScanLine=0;iScanLine<fhb.wHeight;iScanLine++)
    {
        iScanOff = iScanLine*iXOff;
        iY=iYorigin+iScanLine;
        iX=iXorigin;
        // get the raster, add to it , then put it back
        // this is faster than calculating pixels each time...
        VesaRaster(ptr,0,iY,gib.wDefaultWidth,READFLAG);

        for (iByte=0;iByte<iTarget;iByte++)
        {
         for(iXFactor=0;iXFactor<iXOff;iXFactor++)
         {
          iChar = lpFontData[(iYOff*lpszStr[iByte])+(iScanOff+iXFactor)];
          if (iChar != 0)
             for (iBit=0;iBit<8;iBit++)
                 if (iChar & 0X80>>iBit)
                    ptr[iX+(iXFactor*8)+iBit] = wColor;
          }
          // get the proportional spacing info
          // use one pixel to buffer each character
          iWidth = lpFontData[(iYOff*lpszStr[iByte])+iYOff-1];
          iX+= iWidth;
        }
        // put the raster back
        VesaRaster(ptr,0,iY,gib.wDefaultWidth,WRITEFLAG);

      }

return 0;
}

