// =========================================================================
// Program Name : VesaLib.h
// Author       : Bill Buckels
// Date         : September 25, 1995
// Purpose      : Header File
//                Core Graphics Library Routines for VESA 256 Color Mode
// =========================================================================

#include <io.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include <string.h>

// =========================================================================
//                Declarations, Macros, and Constants
// =========================================================================

typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned long ulong;

// global definitions
#define VESA     0x4f
#define VIDEO    0x10
#define VGASEG   0xa000
#define TRUE     1
#define FALSE    0
#define SVGA     256
#define VGA      16
#define MONO     2
#define MAXVGABUF  640
#define MAXSCANBUF MAXVGABUF*3
#define PIXEL    1

#define SVGAPLANES  1
#define SVGABITS    8
#define SVGAWIDTH   640
#define SVGAHEIGHT  480

// status codes
#define SUCCESS     0
#define INITFAIL    1
#define EXECFAIL    2
#define UNSUPPORTED 3
#define UNAVAILABLE 4
#define WRITEFLAG   0
#define READFLAG    1
#define NOSCRIPT   -1
#define EMPTY      -2
#define INVALID    -3
#define NOMEM      -4
#define NOTPCX     99

// action codes
#define BMP         1
#define PCX         2
#define TEXT        3
#define MENU        4
#define LABEL       5
#define DELAY       6

typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned long ulong;

typedef struct{
    char    VesaSignature[4];   // 0: 4 signature bytes:  must be 'VESA'
    uint    VesaVersion;        // 4: VESA version number
    char    far *OEMStringPtr;  // 6: Pointer to OEM string
    long    Capabilities;       // 10: Capabilities of the video environment
    int     *VideoModePtr;      // 14: list of supported Super VGA modes
    char    VIBfiller[256-18];  // 18: fill to 256 bytes
}VgaInfoBlock;

typedef struct{
    uint    ModeAttributes;     // 0: mode attributes
    uchar   WinAAttributes;     // 2: Window A attributes
    uchar   WinBAttributes;     // 3: Window B attributes
    uint    WinGranularity;     // 4: Window granularity
    uint    WinSize;            // 6: Window size
    uint    WinASegment;        // 8: Window A start segment
    uint    WinBSegment;        // 10: Window B start segment
    uint    WinFuncPtr[2];      // 12: (far) pointer to window function
    uint    BytesPerScanLine;   // 14: bytes per scan line

    uint    XResolution;        // 16: horizontal resolution
    uint    YResolution;        // 18: vertical resolution
    uchar   XCharSize;          // 19: character cell width
    uchar   YCharSize;          // 20: character cell height
    uchar   NumberOfPlanes;     // 21: number of memory planes
    uchar   BitsPerPixel;       // 22: bits per pixel
    uchar   NumberOfBanks;      // 23: number of banks
    uchar   MemoryModel;        // 24: memory model type
    uchar   BankSize;           // 25: bank size in KB

    char    MIBfiller[256-26];  // 26: fill to 256 bytes
}ModeInfoBlock;

typedef struct{
    uint  wDefaultWidth;         // 640 or User Override
    uint  wDefaultHeight;        // 480 or User Override
    uint  wDefaultPlanes;        // 1 byte plane
    uint  wDefaultBits;          // 8 bits per pixel
    uint  wVesaMode;             // current mode
    uint  wSaveMode;             // initial mode
    uint  wWriteBank;            // window value for set bank
    uint  wReadBank;             // window value for set bank
    uint  wSaveBank;             // last value from set bank
    uchar far *lpCrt;            // pointer to the screen address
    uchar far *lpScanLine;       // global scan line buffer
    ulong dwBankSize;            // Window Size in Bytes for Current Mode
    uchar aRgbInfo[256][3];      // DAC settings
    uchar aOldRgbInfo[256][3];   // previous DAC settings
    uchar wBlack;                // darkest current color DAC index
    uchar wWhite;                // lightest current color DAC index
    uint  wXorigin;              // xorigin for load point - from left
    uint  wYorigin;              // yorigin for load point - from bottom
    uint  wBg, wFg;              // monochrome color flags
}GlobalInfoBlock;

typedef struct
    {
    char  szFileType[8];     // FREEFONT
    uint  wWidth    ;        // width
    uint  wHeight   ;        // height
    ulong dwHeadSize;        // header size - offset to fontdata
    uint  wBufSize  ;        // buffer size to allocate for font
    }FontHeaderBlock;

// union for breaking-up an intel long into its components
// i.e. the least significant and most significant words
union HILOLONG{
    ulong l;
    uint  x[2];
    };

#define MSB  1
#define LSB  0
#define WSEG 1
#define WOFF 0

// Bitmap Header structures
typedef struct tagRGBQUAD
{
    uchar    rgbBlue;
    uchar    rgbGreen;
    uchar    rgbRed;
    uchar    rgbReserved;
} RGBQUAD;

typedef struct tagBITMAPINFOHEADER
{
    ulong   biSize;
    ulong   biWidth;
    ulong   biHeight;
    uint    biPlanes;
    uint    biBitCount;
    ulong   biCompression;
    ulong   biSizeImage;
    ulong   biXPelsPerMeter;
    ulong   biYPelsPerMeter;
    ulong   biClrUsed;
    ulong   biClrImportant;
} BITMAPINFOHEADER;

// constants for the biCompression field
#define BI_RGB      0L
#define BI_RLE8     1L
#define BI_RLE4     2L

typedef struct tagBITMAPINFO
{
    BITMAPINFOHEADER bmiHeader;
    RGBQUAD          bmiColors[256];
} BITMAPINFO;


typedef struct tagBITMAPFILEHEADER
{
    uchar   bfType[2];
    ulong   bfSize;
    uint    bfReserved1;
    uint    bfReserved2;
    ulong   bfOffBits;
} BITMAPFILEHEADER;

typedef struct{
    int iFormat;
    int iByteBoundary;
    int iPalColors;
    int iBufOff;
    int iDivisor;
}UniversalFormat;

// Header Information for VESALIB2.C

#define RIGHTSHIFT 2
#define LEFTSHIFT  1

#define ENTERKEY   '\x0d' /* char -  the Enter Key          */
#define ESCKEY     '\x1b' /* char -  the Esc key            */
#define FUNCKEY    '\x00' /* first char -  function keys    */
#define UPARROW    'H'    /* second char -  up-arrow key    */
#define DOWNARROW  'P'    /* second char -  down-arrow key  */
#define LEFTARROW  'K'    /* second char -  left-arrow key  */
#define RIGHTARROW 'M'    /* second char -  right-arrow key */

/* a bit field structure */
/* for the poop from from the keystatus */
struct shiftflag{
       unsigned right_shift: 1;
       unsigned left_shift : 1;
       unsigned ctrl       : 1;
       unsigned alt        : 1;
       unsigned scroll_lock: 1;
       unsigned num_lock   : 1;
       unsigned caps_lock  : 1;
       unsigned insert     : 1;
    };

// 128 byte header... note: pack structure on byte boundaries
typedef struct{
    unsigned char zsoft;                // 10
    unsigned char version;              // 5
    unsigned char codetype;             // 1
    unsigned char pixbits;              // 8
    unsigned int  xmin,ymin;
    unsigned int  xmax,ymax;
    unsigned int  hres,vres;
    unsigned char egapalette[3][16];
    unsigned char reserved;
    unsigned char no_planes;            // 1
    unsigned int  bytes_per_line;
    unsigned int  pal_type;
    unsigned char unused[58];
}PCXHEADER;

typedef struct{
    int iTopx;
    int iTopy;
    int iBotx;
    int iBoty;
    unsigned char Pal[769];
    }ScreenWindow;

// draw an xored box
#define XDot(x,y)  *PixAddr(x,y)^='\xff'

// =========================================================================
//                Global Data
// =========================================================================

#ifdef VESALIB
  VgaInfoBlock far vib;
  ModeInfoBlock far mib;
  GlobalInfoBlock gib;
  uint wClsFlag = FALSE;
  union HILOLONG hl;
  union REGS vinregs, voutregs;              // register variables
  struct SREGS vsegregs;                    // for interrupt calls
  // the default vga 16 color palette (8 bit values)
  // start with this
  unsigned char VgaPalette[16][3]={
        0x00, 0x00, 0x00,
        0x00, 0x00, 0xAA,
        0x00, 0xAA, 0x00,
        0x00, 0xAA, 0xAA,
        0xAA, 0x00, 0x00,
        0xAA, 0x00, 0xAA,
        0xAA, 0xAA, 0x00,
        0xAA, 0xAA, 0xAA,
        0x55, 0x55, 0x55,
        0x55, 0x55, 0xFF,
        0x55, 0xFF, 0x55,
        0x55, 0xFF, 0xFF,
        0xFF, 0x55, 0x55,
        0xFF, 0x55, 0xFF,
        0xFF, 0xFF, 0x55,
        0xFF, 0xFF, 0xFF};
  UniversalFormat uif;
  BITMAPFILEHEADER BitMapFileHeader;
  BITMAPINFO       bmp;
  FontHeaderBlock fhb;
  uchar *lpFontData;
  uint static wFontFlag=FALSE;
#else
  extern VgaInfoBlock far vib;
  extern ModeInfoBlock far mib;
  extern GlobalInfoBlock gib;
  extern uint wClsFlag;
  extern union HILOLONG hl;
  extern union REGS vinregs, voutregs;
  extern struct SREGS vsegregs;
  extern unsigned char VgaPalette[16][3];
  extern UniversalFormat uif;
  extern BITMAPFILEHEADER BitMapFileHeader;
  extern BITMAPINFO       bmp;
  extern FontHeaderBlock fhb;
  extern uchar *lpFontData;
  extern uint wFontFlag;
#endif

#ifdef VESALIB2
   // the key status is at this address
   char far *keystatus= (char far *)(unsigned long)0x00400017;

   // we use a union to unmask the bit fields
   union keyboard{
      struct shiftflag bit;
      unsigned char status;
    }flag;

   PCXHEADER pcxheader;
   ScreenWindow sw;
#else
   extern char far *keystatus;
   extern PCXHEADER pcxheader;
   extern ScreenWindow sw;
#endif

// =========================================================================
//                Function Prototypes
// =========================================================================

/* vesalib.c 24/09/95 07.16.48 */
int GetVgaInfo (void);
int GetModeInfo (uint wVesaMode);
uint FindVesaMode (uint xres, uint vres, uint noplanes, int pixelbits);
uint SetVesaMode (uint wVesaMode);
uint GetVesaMode (void);
void SetVesaBank (uint wWin, uint wBank);
int InitVesaSession (void);
void LoadPalette (void);
unsigned char far *PixAddr (uint x, uint y);
unsigned char ReadPixel (uint x, uint y);
void WritePixel (uint x, uint y, uint color);
void VLine (int y1, int y2, int x, unsigned color);
void HLine (int x1, int x2, int y, unsigned color);
int DrawLine (int x1, int y1, int x2, int y2, unsigned color);
void DrawBox (int x1, int y1, int x2, int y2, unsigned color);
void GetAspect (int *x, int *y);
void DrawCircle (int cx, int cy, int r, int color, int fillcolor);
void MoveBytes (char far *lpTarget, char far *lpSource, uint wBytes);
void VesaRaster (uchar *lpszBuf,
	 uint x,
	 uint y,
	 uint wPixels,
	 uint wFLAG);
void FBox (uint x1, uint y1, uint x2, uint y2, uchar color);
void ClearScreen (uint color);
void Vga2Svga (unsigned char *src, unsigned char *dest, unsigned width);
void Mono2Svga (unsigned char *src,
	 unsigned char *dest,
	 unsigned width);
int LoadVesaBitMap (char *lpszBmpFile);
void FreeVesaFont (void);
int LoadVesaFont (char *lpszFontName);
int DrawVesaFont (uchar *lpszStr,
	 int iXorigin,
	 int iYorigin,
	 char cJustify,
	 uint wColor);

/* vesalib2.c 27/09/95 13.16.26 */
int shiftstatus (void);
void HXLine (uint x1, uint x2, uint y);
void VXLine (unsigned y1, unsigned y2, unsigned x);
void XBox (unsigned x1, unsigned y1, unsigned x2, unsigned y2);
void XBlox (unsigned x1, unsigned y1, unsigned x2, unsigned y2);
int PcxPut (FILE *PcxFile, unsigned char byt, unsigned char cnt);
int WritePcxLine (FILE *PcxFile, unsigned char *ptr, int bytes);
int SavePcx (char *lpszClipName);
void ShowPcx (FILE *PcxFile,
	 unsigned y1,
	 unsigned y2,
	 unsigned x1,
	 unsigned bytes);
int LoadPcx (char *lpszClipName);

#ifndef MOUSELIB
  extern int mouse;
#endif

#define TIMEOUT 0x2c00
#define SILENCE 32767

/* mouselib.c 14/11/95 00.32.58 */
int sound (int freq, int tlen);
void keyclick (void);
void bronx (void);
void bell (void);
int mousebutton (void);
int initmouse (void);
